% Learns classifier and classifies test set
%  by representing the distributions as Gaussians, 
%  with diagonal covariance matrix
%  learning the parameters using Maximum Likelihood
%  
% Inputs:
% Usage
%      [trainError, testError, estTrainLabels, estTestLabels] = ...
%           ML_diag_VC(trainFeatures, trainLabels,unused ,testFeatures, testLabels)
% where
%
% Inputs:
% 	trainFeatures  - the training set vectors, one vector per column
%	trainLabels    - the labels of the above
%       unuse          - unused parameter
%       testFeatures   - test set, one column per vector
%       testLabels     - labels  for test set%
% Outputs
%	trainError     - the error rate on the training set (one entry per
%	                   class + total error)
%	testError      - the error rate on the test set (one entry per class
%	                  + total error)
%       estTrainLabels - the labels produced by the algorithm for the
%                          training samples
%       estTestLabels - the labels produced by the algorithm for the
%                          test samples


function [trainError, testError, estTrainLabels, estTestLabels] =  ...
    ML_diag_VC(trainFeatures, trainLabels, AlgorithmParameters, testFeatures, testLabels);

[Nclasses, classes]  = find_classes([trainLabels(:);testLabels(:)]); % Number of classes in labels
[Dim, Ntrain]           = size(trainFeatures);
[Dim, Ntest]            = size(testFeatures);

for cl =1:Nclasses,
  trainIndexes   = find(trainLabels == classes(cl));
  priors(cl)     = length(trainIndexes);
  means(:,cl)    = (mean(trainFeatures(:,trainIndexes)'))';
  vars(:,cl)     = var(trainFeatures(:,trainIndexes)')';
end
priors = priors/length(trainLabels);

% now classifies
% first training set
counts = zeros(Nclasses, Ntrain);
oneVect = ones(1,Ntrain);
for (cl = 1:Nclasses)
  covariance = vars(:,cl);
  determinant = prod(covariance);
  covariance  = diag(covariance,0);
  covariance  = inv(covariance);
  meanval    = means(:,cl) * oneVect;
  diff       = trainFeatures - meanval;
  
  diff       = exp(-sum((diff' * covariance)'.*diff)/2);
  counts(cl,:) = diff/sqrt(determinant) * priors(cl);
end

[counts, indexes] = max(counts);
estTrainLabels = classes(indexes);

% then test set
counts = zeros(Nclasses, Ntest);
oneVect = ones(1,Ntest);
for (cl = 1:Nclasses)
  covariance = vars(:,cl);
  determinant = prod(covariance);
  covariance = covariance.^(-1);
  covariance = diag(covariance);
  meanval    = means(:,cl) * oneVect;
  diff       = testFeatures - meanval;
  
  diff       = exp(-sum((diff' * covariance)'.*diff)/2);
  counts(cl,:) = diff/sqrt(determinant) * priors(cl);
end

[counts, indexes] = max(counts);
estTestLabels = classes(indexes);

trainError = computeError(classes, trainLabels, estTrainLabels);
testError  = computeError(classes, testLabels , estTestLabels);

